from pyodide.ffi import create_proxy
from js import document as DOC
from functools import partial
from typing import assert_never


def vars_pattern(X: str):
    return " & ".join(X.split())


patterns = {
    "vars": {
        "desc": "variables in X are True",
        "tpl": vars_pattern,
        "args": [{"name": "X", "type": "text", "desc": "space-separated list"}],
    },
    "form": {
        "desc": "formula X holds",
        "tpl": "{X}",
        "args": [{"name": "X", "type": "text", "desc": "CTL formula"}],
    },
    "and": {
        "desc": "X and Y hold",
        "tpl": "({X}) & ({Y})",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
    "or": {
        "desc": "X or Y hold",
        "tpl": "({X}) | ({Y})",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
    "not": {
        "desc": "X does not hold",
        "tpl": "~({X})",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "reach-may": {
        "desc": "X may be reached",
        "tpl": "EF({X})",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "reach-must": {
        "desc": "X must be reached",
        "tpl": "AF({X})",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "inv-may-persist": {
        "desc": "X may persist forever",
        "tpl": "EG({X})",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "inv-must-persist": {
        "desc": "X must persist forever",
        "tpl": "AG({X})",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "inv-may-reach": {
        "desc": "X may remain forever reachable",
        "tpl": "EG(EF({X}))",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "inv-must-reach": {
        "desc": "X must remain forever reachable",
        "tpl": "AG(EF({X}))",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "inv-infinite-reach": {
        "desc": "X is necessarily reached infinitely often",
        "tpl": "AG(AF({X}))",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "reach-inv-may": {
        "desc": "is is possible to reach a state from which X may persist forever",
        "tpl": "EF(EG({X}))",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "reach-inv-must": {
        "desc": "is is possible to reach a state from which X must persist forever",
        "tpl": "EF(AG({X}))",
        "args": [{"name": "X", "type": "pattern"}],
    },
    "then-may": {
        "desc": "if X is reached, then it is possibly followed by Y",
        "tpl": "AF(({X}) => EF({Y}))",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
    "then-must": {
        "desc": "if X is reached, then it is necessarily followed by Y",
        "tpl": "AF(({X}) => EF({Y}))",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
    "seq-may-sometimes": {
        "desc": "Y is reachable and is possibly preceded at some time by X",
        "tpl": "EF(({X}) & EF({Y}))",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
    "seq-may-always": {
        "desc": "Y is reachable and is possibly preceded all the time by X",
        "tpl": "E(({X}) U ({Y}))",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
    "seq-must-sometimes": {
        "desc": "Y is reachable and is necessarily preceded at some time by X",
        "tpl": "EF({X}) & (~E((~({X})) U ({Y})))",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
    "seq-must-always": {
        "desc": "Y is reachable and is necessarily preceded all the time by X",
        "tpl": "EF({X}) & AG((~({X})) => AG(~({Y})))",
        "args": [{"name": "X", "type": "pattern"}, {"name": "Y", "type": "pattern"}],
    },
}


class HTML:
    def __call__(self, tag, *content, **attrs):
        elt = DOC.createElement(tag)
        for key, val in attrs.items():
            elt.setAttribute(key, val)
        for child in content:
            if isinstance(child, str):
                elt.innerHTML = elt.innerHTML + child
            else:
                elt.appendChild(child, append=True)
        return elt

    def __getattr__(self, tag):
        return partial(self.__call__, tag)


H = HTML()
WIZ = DOC.querySelector("#wiz")


class WizUI:
    def __init__(self):
        div = H.div(
            (form := H.div(CLASS="wiz-form")),
            H.div(
                (out := H.textarea(name="output", readonly=True)),
                (button := H.button("Copy")),
                CLASS="wiz-output",
            ),
            CLASS="wiz-wizard",
        )
        self.child = PatternUI(self, form)
        self.output = out
        WIZ.replaceChild(div, WIZ.firstChild)
        button.addEventListener("click", create_proxy(self.on_copy))

    def update(self):
        self.output.value = self.child.formula()
        self.output.setSelectionRange(0, 0)

    def on_copy(self, *_):
        self.output.select()
        self.output.setSelectionRange(0, 99999)
        DOC.execCommand("copy")


class PatternUI:
    def __init__(self, wiz, container):
        self.wiz = wiz
        pat = H.div(
            H.div(
                H.label("Pattern:", FOR="pattern"),
                (
                    select := H.select(
                        *(
                            H.option(f"{spec['desc']} [{name}]", value=name)
                            for name, spec in patterns.items()
                        ),
                        name="pattern",
                    )
                ),
                CLASS="wiz-select",
            ),
            (args := H.div(CLASS="wiz-args")),
            CLASS="wiz-pattern",
        )
        self.select, self.args = select, args
        container.appendChild(pat)
        self.select.addEventListener("change", create_proxy(self.on_select))
        self.on_select()

    def on_select(self, *args):
        val = str(self.select.value)
        self.mk(val)
        if args:
            self.wiz.update()

    def mk(self, name):
        self.children = {}
        spec = self.pattern = patterns[name]
        self.args.innerHTML = ""
        for arg in spec["args"]:
            if arg["type"] == "text":
                elt = H.div(
                    (inp := H.input(TYPE="text", placeholder=arg["desc"])),
                    CLASS="wiz-arg-value",
                )
                self.children[arg["name"]] = inp
                elt.addEventListener("input", create_proxy(self.on_update))
            elif arg["type"] == "pattern":
                elt = H.div(CLASS="wiz-arg-value")
                self.children[arg["name"]] = PatternUI(self.wiz, elt)
            else:
                assert_never(arg)
            div = H.div(
                H.div(H.label(f"{arg['name']}:"), CLASS="wiz-arg-name"),
                elt,
                CLASS="wiz-arg",
            )
            self.args.appendChild(div, append=True)

    def on_update(self, *_):
        self.wiz.update()

    def formula(self):
        tpl = self.pattern["tpl"]
        args = {}
        for key, val in self.children.items():
            if isinstance(val, PatternUI):
                args[key] = val.formula()
            else:
                args[key] = val.value
        if isinstance(tpl, str):
            return tpl.format(**args)
        else:
            return tpl(**args)


WizUI()
